#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __main__.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""isbg scans an IMAP Inbox and runs every entry against SpamAssassin.

See Also:
    :py:mod:`isbg` for instance creation.

.. versionchanged:: 2.0
    ``--partialrun`` defaults to 50. Use ``--partialrun=0`` to run without
    partial run.

"""

import os
import sys

try:
    # Creating command-line interface
    from docopt import docopt, DocoptExit, printable_usage
except ImportError:
    sys.stderr.write("Missing dependency: docopt\n")
    raise

if __package__ is None and not hasattr(sys, 'frozen'):
    # direct call of __main__.py
    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))
from isbg import isbg  # noqa: E402


def __cmd_opts__():  # noqa: D207
    """Isbg scans an IMAP Inbox and runs every entry against SpamAssassin.

Command line Options::

 Usage:
  isbg.py --imaphost <hostname> --imapuser <username> [options]
  isbg.py --imaphost <hostname> --imapuser <username> --imaplist [options]
  isbg.py (-h | --help)
  isbg.py --usage
  isbg.py --version

 Options:
  --imaphost hostname    IMAP server name.
  --imapuser username    Who you login as.

  --imaplist             List imap directories.

  -h, --help             Show the help screen.
  --usage                Show the usage information.
  --version              Show the version information.

  --dryrun               Do not actually make any changes.
  --delete               The spams will be marked for deletion from
                         your inbox.
  --deletehigherthan #   Delete any spam with a score higher than #.
  --exitcodes            Use exitcodes to detail  what happened.
  --expunge              Cause marked for deletion messages to also be
                         deleted (only useful if --delete is
                         specified).
  --flag                 The spams will be flagged in your inbox.
  --gmail                Delete by copying to '[Gmail]/Trash' folder.
  --ignorelockfile       Don't stop if lock file is present.
  --imappasswd passwd    IMAP account password.
  --imapport port        Use a custom port.
  --imapinbox mbox       Name of your inbox folder [Default: INBOX].
  --learnspambox mbox    Name of your learn spam folder.
  --learnhambox mbox     Name of your learn ham folder.
  --learnthendestroy     Mark learnt messages for deletion.
  --learnthenflag        Flag learnt messages.
  --learnunflagged       Only learn if unflagged
                         (for  --learnthenflag).
  --learnflagged         Only learn flagged.
  --lockfilegrace=<min>  Set the lifetime of the lock file
                         [default: 240.0].
  --lockfilename file    Override the lock file name.
  --maxsize numbytes     Messages larger than this will be ignored as
                         they are unlikely to be spam.
  --movehamto mbox       Move ham to folder.
  --noninteractive       Prevent interactive requests.
  --noreport             Don't include the SpamAssassin report in the
                         message copied to your spam folder.
  --nostats              Don't print stats.
  --partialrun num       Stop operation after scanning 'num' unseen
                         emails. Use 0 to run without partial run
                         [default: 50].
  --passwdfilename fn    Use a file to supply the password.
  --savepw               Store the password to be used in future runs.
  --spamc                Use spamc instead of standalone SpamAssassin
                         binary.
  --spaminbox mbox       Name of your spam folder
                         [Default: INBOX.Spam].
  --nossl                Don't use SSL to connect to the IMAP server.
  --teachonly            Don't search spam, just learn from folders.
  --trackfile file       Override the trackfile name.
  --verbose              Show IMAP stuff happening.
  --verbose-mails        Show mail bodies (extra-verbose).

  (Your inbox will remain untouched unless you specify --flag or
   --delete)
    """


def parse_args(sbg):
    """Argument processing of the command line.

    :param sbg: the `isbg.ISBG` instance which would be updated with the
                parameters.
    :type sbg: isbg.ISBG
    :return: `None`

    :Example: You can run it using:

        >>> sbg = isbg.ISBG()
        >>> parse_args(sbg)
    """
    try:
        opts = docopt(__cmd_opts__.__doc__, version="isbg_v" +
                      isbg.__version__ + ", from: " +
                      os.path.abspath(__file__) + "\n\n" + isbg.__license__)
        opts = dict([(k, v) for k, v in opts.items()
                     if v is not None])
    except DocoptExit as exc:
        raise isbg.ISBGError(isbg.__exitcodes__['flags'],
                             "Option processing failed - " + str(exc))

    if opts.get("--usage"):
        print(printable_usage(__cmd_opts__.__doc__))
        return 1

    if opts.get("--deletehigherthan") is not None:
        try:
            sbg.deletehigherthan = float(opts["--deletehigherthan"])
        except Exception:  # pylint: disable=broad-except
            raise isbg.ISBGError(isbg.__exitcodes__['flags'],
                                 "Unrecognized score - " +
                                 opts["--deletehigherthan"])
        if sbg.deletehigherthan < 1:
            raise isbg.ISBGError(isbg.__exitcodes__['flags'],
                                 "Score " + repr(sbg.deletehigherthan) +
                                 " is too small")

    if opts["--flag"] is True:
        sbg.spamflags.append("\\Flagged")

    sbg.imapsets.host = opts.get('--imaphost', sbg.imapsets.host)
    sbg.imapsets.passwd = opts.get('--imappasswd',
                                   sbg.imapsets.passwd)
    sbg.imapsets.port = opts.get('--imapport', sbg.imapsets.port)
    sbg.imapsets.user = opts.get('--imapuser', sbg.imapsets.user)
    sbg.imapsets.inbox = opts.get('--imapinbox', sbg.imapsets.inbox)
    sbg.imapsets.spaminbox = opts.get('--spaminbox', sbg.imapsets.spaminbox)
    sbg.imapsets.learnspambox = opts.get('--learnspambox')
    sbg.imapsets.learnhambox = opts.get('--learnhambox')
    sbg.imapsets.nossl = opts.get('--nossl', sbg.imapsets.nossl)

    sbg.lockfilegrace = float(opts.get('--lockfilegrace', sbg.lockfilegrace))

    sbg.nostats = opts.get('--nostats', False)
    sbg.dryrun = opts.get('--dryrun', False)
    sbg.delete = opts.get('--delete', False)
    sbg.gmail = opts.get('--gmail', False)

    if opts.get("--maxsize") is not None:
        try:
            sbg.maxsize = int(opts["--maxsize"])
        except (TypeError, ValueError):
            raise isbg.ISBGError(isbg.__exitcodes__['flags'],
                                 "Unrecognised size - " + opts["--maxsize"])
        if sbg.maxsize < 1:
            raise isbg.ISBGError(isbg.__exitcodes__['flags'],
                                 "Size " + repr(sbg.maxsize) + " is too small")

    sbg.movehamto = opts.get('--movehamto')

    if opts["--noninteractive"] is True:
        sbg.interactive = 0

    sbg.noreport = opts.get('--noreport', sbg.noreport)

    sbg.lockfilename = opts.get('--lockfilename', sbg.lockfilename)

    sbg.trackfile = opts.get('--trackfile', sbg.trackfile)

    sbg.partialrun = opts.get('--partialrun', sbg.partialrun)
    try:
        sbg.partialrun = int(opts["--partialrun"])
    except ValueError:
        raise isbg.ISBGError(isbg.__exitcodes__['flags'],
                             "partialrun \'{}\' must be a integer".format(
                             repr(sbg.partialrun)))
    if sbg.partialrun < 0:
        raise isbg.ISBGError(isbg.__exitcodes__['flags'],
                             ("Partial run \'{}\' number must be equal to " +
                              "0 or higher").format(repr(sbg.partialrun)))
    elif sbg.partialrun == 0:
        sbg.partialrun = None

    sbg.verbose = opts.get('--verbose', sbg.verbose)
    sbg.verbose_mails = opts.get('--verbose-mails', sbg.verbose_mails)
    sbg.ignorelockfile = opts.get("--ignorelockfile", sbg.ignorelockfile)
    sbg.savepw = opts.get('--savepw', sbg.savepw)
    sbg.passwdfilename = opts.get('--passwdfilename', sbg.passwdfilename)

    sbg.imaplist = opts.get('--imaplist', sbg.imaplist)

    sbg.learnunflagged = opts.get('--learnunflagged', sbg.learnunflagged)
    sbg.learnflagged = opts.get('--learnflagged', sbg.learnflagged)
    sbg.learnthendestroy = opts.get('--learnthendestroy', sbg.learnthendestroy)
    sbg.learnthenflag = opts.get('--learnthendestroy', sbg.learnthenflag)
    sbg.expunge = opts.get('--expunge', sbg.expunge)

    sbg.teachonly = opts.get('--teachonly', sbg.teachonly)
    sbg.spamc = opts.get('--spamc', sbg.spamc)

    sbg.exitcodes = opts.get('--exitcodes', sbg.exitcodes)

    # fixup any arguments
    if opts.get("--imapport") is None:
        if opts["--nossl"] is True:
            sbg.imapsets.port = 143
        else:
            sbg.imapsets.port = 993


def main():
    """Run when this module is called from the command line.

    When the main function ends, it throw a sys.exit with 0 if it has end ok
    or one of the :py:data:`isbg.isbg.__exitcodes__`
    """
    sbg = isbg.ISBG()
    try:
        if parse_args(sbg) == 1:  # usage option
            sys.exit(0)
        return sbg.do_isbg()  # return the exit code.
    except isbg.ISBGError as err:
        sys.stderr.write(err.message)
        sys.stderr.write("\nUse --help to see valid options and arguments\n")
        if err.exitcode == -1:
            raise
        sys.exit(err.exitcode)


if __name__ == '__main__':
    isbgret = main()  # pylint: disable=invalid-name
    if isbgret is not None:
        sys.exit(isbgret)



# -*- coding: utf-8 -*-
#
"""Conf file for isbg docs using sphinx."""

import re
from ast import literal_eval
import os
import sys
import recommonmark


# -- Custom variables -----------------------------------------------------

cvar_gitlab_base_uri = 'https://gitlab.com/'
cvar_gitlab_prj = 'isbg'
cvar_gitlab_usr = 'isbg'
cvar_gitlab_uri = cvar_gitlab_base_uri + cvar_gitlab_prj + '/' + \
    cvar_gitlab_usr
cvar_pypi_uri = 'https://pypi.python.org/pypi/isbg'

master_doc = 'index'  # The master toctree document.

project = u'isbg'
copyright = u'''License GPLv3: GNU GPL version 3 https://gnu.org/licenses/gpl.html

This is free software: you are free to change and redistribute it. There is
NO WARRANTY, to the extent permitted by law.'''

author = u'''See the CONTRIBUTORS file in the git repository for more
information on who wrote and maintains this software'''

# We get the version from isbg/isbg.py
_VERSION_RE = re.compile(r'__version__\s+=\s+(.*)')
with open('../isbg/isbg.py', 'rb') as f:
    _VERSION = str(literal_eval(_VERSION_RE.search(
        f.read().decode('utf-8')).group(1)))

sys.path.insert(0, os.path.abspath('..'))


# -- Extensions -----------------------------------------------------------

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.intersphinx',
              'sphinx.ext.napoleon',
              'sphinx.ext.coverage',
              'sphinx.ext.mathjax',
              'sphinx.ext.ifconfig',
              'sphinx.ext.extlinks',
              'sphinx.ext.autosummary',
              'sphinx.ext.todo'
              ]

source_suffix = ['.rst']

version = _VERSION
release = _VERSION

pygments_style = 'sphinx'


# -- Options for todo extension -------------------------------------------

todo_include_todos = True


# -- Options for autodoc extension ----------------------------------------

autodoc_member_order = 'bysource'
autodoc_default_flags = ['members']

autodoc_docstring_signature = False
autodoc_mock_imports = []
autodoc_warningiserror = True
autodoc_default_options = {'exclude-members': 'xdg_cache_home'}

# Enable nitpicky mode - which ensures that all references in the docs
# resolve.

nitpicky = True
nitpick_ignore = []

for line in open('nitpick-exceptions'):
    if line.strip() == "" or line.startswith("#"):
        continue
    dtype, target = line.split(None, 1)
    target = target.strip()
    nitpick_ignore.append((dtype, target))


# -- Options for napoleon extension ---------------------------------------

napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True


# -- Options for extlinks extension ---------------------------------------

extlinks = {
    'issue': (cvar_gitlab_uri + '/issues/%s', 'issue '),  # e.g. :issue:`12`
    'pull': (cvar_gitlab_uri + '/pull/%s', 'pull ')      # e.g. :pull:`11`
}


# -- Options for HTML output ----------------------------------------------

html_theme = 'sphinx_rtd_theme'

# For theme 'sphinx_rtd_theme':
html_theme_options = {
    'canonical_url': '',
    'analytics_id': '',
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'collapse_navigation': False,
    'sticky_navigation': True,
    'navigation_depth': 3,
}

html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'searchbox.html',
        'relations.html',
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

htmlhelp_basename = 'isbgdoc'


# -- Options for manual page output ---------------------------------------

man_pages = [
    ('manpage.isbg', 'isbg', u'scans an IMAP Inbox and runs every entry ' +
     u'against SpamAssassin.',
     [author], 1),
    ('manpage.isbg_sa_unwrap', 'isbg_sa_unwrap', u'unwraps a email bundeled ' +
     u'by SpamAssassin.',
     [author], 1),
]

intersphinx_mapping = {'https://docs.python.org/': None}


# -- Generate the documentation from sources -----

def run_apidoc(_):
    """Run apidoc."""
    from sphinx.ext.apidoc import main
    import os
    import sys
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    module = os.path.join(cur_dir, "..", project)
    params = ['-e', '--force', '--separate', '--private', '--follow-links',
              '-o', cur_dir, module]
    main(params)


def import_rsts(_):
    """Copy rst files from base dir to cur dir."""
    import glob
    import shutil
    import os
    import sys
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    for file in glob.glob('../*.rst'):
        shutil.copy2(file, cur_dir)


def import_mds(_):
    """Copy md files from base dir to cur dir."""
    import glob
    import shutil
    import os
    import sys
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    for file in glob.glob('../*.md'):
        shutil.copy2(file, cur_dir)


def setup(app):
    """Configure sphinx."""
    app.connect('builder-inited', run_apidoc)
    app.connect('builder-inited', import_rsts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  test_isbg.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""Test cases for isbg module."""

# With atexit._run_exitfuncs()  we free the lockfile, but we lost coverage
# statistics.

import os
import sys
try:
    import pytest
except ImportError:
    pass

# We add the upper dir to the path
sys.path.insert(0, os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..')))
from isbg import isbg  # noqa: E402


def test_ISBGError():
    """Test a ISBGError object creation."""
    with pytest.raises(isbg.ISBGError, match="foo"):
        raise isbg.ISBGError(0, "foo")


class TestISBG(object):
    """Tests for class ISBG."""

    def test_set_filename(self):
        """Test set_filename."""
        sbg = isbg.ISBG()
        filename = isbg.ISBG.set_filename(sbg.imapsets, "track")
        assert os.path.dirname(filename) != ""
        assert os.path.basename(filename) != ""
        assert os.path.basename(filename).startswith("track")
        filename = isbg.ISBG.set_filename(sbg.imapsets, "password")
        assert os.path.dirname(filename) != ""
        assert os.path.basename(filename) != ""
        assert os.path.basename(filename).startswith(".isbg-")

    def test_removelock(self):
        """Test removelock."""
        sbg = isbg.ISBG()
        sbg.removelock()
        assert os.path.exists(sbg.lockfilename) is False, \
            "File should not exist."
        lockfile = open(sbg.lockfilename, 'w')
        lockfile.write(repr(os.getpid()))
        lockfile.close()
        assert os.path.exists(sbg.lockfilename), "File should exist."
        sbg.removelock()
        assert os.path.exists(sbg.lockfilename) is False, \
            "File should not exist."

    def test_do_isbg(self):
        """Test do_isbg."""
        sbg = isbg.ISBG()
        with pytest.raises(isbg.ISBGError, match="specify your imap password"):
            sbg.do_isbg()
            pytest.fail("It should rise a specify imap password " + "ISBGError")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
isbg scans an IMAP Inbox and runs every entry against SpamAssassin.

For any entries that match, the message is copied to another folder, and the
original marked or deleted.

.. autodata:: __version__

.. autodata:: __exitcodes__
    :annotation: = {'ok': 0, ...}

.. autodata:: xdg_cache_home
    :annotation: = /home/user/.cache

"""

import os
import sys     # Because sys.stderr.write() is called bellow

from isbg import imaputils
from isbg import secrets
from isbg import spamproc
from isbg import utils

from .utils import __

import atexit
import getpass
import json
import logging
import re
import time

# xdg base dir specification (only xdg_cache_home is used)
try:
    from xdg.BaseDirectory import xdg_cache_home
except ImportError:
    xdg_cache_home = ""    # pylint: disable=invalid-name
if xdg_cache_home == "":
    # pylint: disable=invalid-name
    xdg_cache_home = os.path.expanduser("~" + os.sep + ".cache")
    """str: From the `XDG Base Directory specification`_.

We used this directory to create a `isbg/` one to store cached data:
    * lock file.
    * password file.
    * cached lists of ``uids``.

Note:
    For best `xdg_cache_home` detection, install *python-xdg*.

.. _XDG Base Directory specification:
    https://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html
"""


__version__ = "2.3.1"  #: The current isbg version

__license__ = \
    """License GPLv3: GNU GPL version 3 https://gnu.org/licenses/gpl.html

This is free software: you are free to change and redistribute it. There is NO
WARRANTY, to the extent permitted by law."""

__exitcodes__ = {
    'ok': 0,
    'newmsgs': 1,
    'newspam': 2,
    'newmsgspam': 3,
    'flags': 10,
    'imap': 11,
    'spamc': 12,
    'tty': 20,
    'locked': 30,
    'error': -1
}
"""Exit codes used as return in the command line usage in
:py:func:`sys.exit`.

================ === ========= ===============================================
Key              Val cmd? [1]_ Description
================ === ========= ===============================================
``'ok'``           0    no     All has finished ok.
``'newmsgs'``      1   yes     There was at least one new message, and none of
                               them were spam.
``'newspam'``      2   yes     There was at least one new message,
                               and all them were spam.
``'newmsgspam'``   3   yes     There were new messages, with at least one spam
                               and one non-spam.
``'flags'``       10    no     Error with the command line options.
``'imap'``        11    no     The IMAP server reported an error or error
                               with the IMAP connection.
``'spamc'``       12    no     Error with spamc/spamassassin call.
``'tty'``         20    no     The program was not launched in an interactive
                               terminal.
``'locked'``      30    no     Error with the *lock* file, is another instance
                               of ``isbg`` must be running.
``'error'``       -1    no     Other errors.
================ === ========= ===============================================

.. [1] ``--exitcodes`` required.
"""


class ISBGError(Exception):
    """Class for the ISBG exceptions.

    Attributes:
        exitcode (int): The exit code for the error.
        message (str): The human readable error message.

    """

    def __init__(self, exitcode=0, message=""):
        """Initialize a ISBGError object.

        Args:
            exitcode (int): Exit code. It must be a
                :py:const:`~isbg.isbg.__exitcodes__` valid value.
            message (str): Human readable error message.
        Raises:
            ValueError: If the `exitcode` is not in `__exitcodes__`.

        """
        self.exitcode = exitcode
        self.message = message

        Exception.__init__(self, message)
        if exitcode not in __exitcodes__.values():
            raise ValueError


class ISBG(object):
    """Main ISBG class.

    See also:
        :py:mod:`isbg.__main__` for command line usage.

    Examples:
        >>> import isbg
        >>> sbg = isbg.ISBG()
        >>> sbg.imapsets.host = "imap.example.org"
        >>> sbg.imapsets.port = 993
        >>> sbg.imapsets.user = "example@example.org"
        >>> sbg.imapsets.passwd = "xxxxxxxx"
        >>> sbg.imapsets.inbox = "INBOX"
        >>> sbg.imapsets.spaminbox = "INBOX.Spam"
        >>> sbg.imapsets.learnspambox = "INBOX.Spam"
        >>> sbg.imapsets.learnhambox = "NOSPAM"
        >>> # Set the number of mails to chech
        >>> sbg.partialrun = 4        # Only check 4 mails for every proc.
        >>> sbg.verbose = True        # Show more info
        >>> sbg.ignorelockfile = True # Ignore lock file
        >>> sbg.removelock()          # if there is a lock file
        >>> sbg.do_isbg()      # Connects to the imap and checks for spam

    Attributes:
        imap (isbg.imaputils.IsbgImap4): class that take care of connection
            and communication with the `IMAP` server. It's initialized calling
            :py:meth:`do_imap_login` and every time that calling
            :py:meth:`do_isbg`.
        imapsets (isbg.imaputils.ImapSettings): Object to store the `IMAP`
            settings. It's initialized when `ISBG` is initialized and also
            stores the IMAP folders used by ISBG.

            It also stores the command line args for:
                :py:attr:`~isbg.imaputils.ImapSettings.user` (imapuser),
                :py:attr:`~isbg.imaputils.ImapSettings.passwd` (imappasswd),
                :py:attr:`~isbg.imaputils.ImapSettings.host` (imaphost),
                :py:attr:`~isbg.imaputils.ImapSettings.port` (imapport),
                :py:attr:`~isbg.imaputils.ImapSettings.nossl` (nossl),
                :py:attr:`~isbg.imaputils.ImapSettings.inbox` (imapinbox),
                :py:attr:`~isbg.imaputils.ImapSettings.spaminbox` (spaminbox),
                :py:attr:`~isbg.imaputils.ImapSettings.learnspambox`
                (learnspambox) and
                :py:attr:`~isbg.imaputils.ImapSettings.learnhambox`
                (learnhambox).

        logger (logging.Logger): Object used to output info. It's initialized
            when `ISBG` is initialized.

    These are attributes derived from the command line and needed for normal
    operations:

    Attributes:
        exitcodes (bool): If True returns more exit codes. Defaults to
            ``True``.
        imaplist (bool): If True shows the folder list. Default to ``False``.
        noreport (bool): If True not adds SpamAssassin report to mails.
            Default to ``False``.
        nostats (bool): If True no shows stats. Default to ``False``.
        verbose_mails (bool): If True shows the email content. Default to
            ``False``.
        verbose: a property that if it's set to True show more information.
            Default to ``False``.

    These are attributes derived for the command line, and needed for
    `SpamAssassin` operations:

    Attributes:
        dryrun (bool): If True don't do changes in the IMAP account. Default to
            ``False``.
        maxsize (int): Max file size to process. Default to ``120,000``.
        teachonly (bool): If True don't search spam, only learn. Default to
            ``False``.
        spamc (bool): If True use spamc instead of standalone SpamAssassin.
            Default to ``False``.
        gmail (bool): If True Delete by copying to `[Gmail]/Trash` folder.
            Default to ``False``.
        deletehigherthan (float): If it's not None, the minimum score from a
            mail to be deleted. Default to ``None``.
        delete (bool): If True the spam mails will be marked for deletion.
            Default to ``False``.
        expunge (bool): If True causes marked for deletion messages to also
            be deleted (only useful if `deleted` is True. Default to ``None``.
        flag (bool): If True the spams will be flagged in your INBOX. Default
            to ``False``.
        learnflagged (bool): If True only learn flagged messages. Default to
            ``False``.
        learnunflagged (bool): If True only learn unflagged messages. Default
            to ``False``.
        learnthendestroy (bool): If True mark learned messages for deletion.
            Default to ``False``.
        learnthenflag (bool): If True flag learned messages. Default to
            ``False``.
        movehamto (str): If it's not None, IMAP folder where the ham mail will
            be moved. Default to ``None``.

    These are attributes derived from the command line and related to the lock
    file:

    Attributes:
        ignorelockfile (bool): If True and there is the lock file a error is
            raised.
        lockfile (str): Full path and name of the lock file.

            The path it's initialized with the xdg cache home specification
            plus `/isbg/` and with the name ``lock``.

        lockfilegrace (float): Lifetime of the lock file in seconds. Default to
            240.0

    These are attributes derived for the command line, related to the
    `IMAP` password and files:

    Attributes:
        passwdfilename (str): The fill name where the password will be stored.
            Defaults to ``None``. It only have use if `savepw` is `True`.
        savepw (bool): If True a obfuscated password will be stored into a
            file.
        trackfile (str): Base name where the processed ``uids`` will be stored
            to not reprocess them. Default to ``None`` when initialized and
            initialized the first time that is needed.

    """

    def __init__(self):
        """Initialize a ISBG object."""
        self.imapsets = imaputils.ImapSettings()
        self.imap = None

        self.logger = logging.getLogger(__name__)       #: a logger
        self.logger.addHandler(logging.StreamHandler())

        # We create the dir for store cached information (if needed)
        if not os.path.isdir(os.path.join(xdg_cache_home, "isbg")):
            os.makedirs(os.path.join(xdg_cache_home, "isbg"))

        self.imaplist, self.nostats = (False, False)
        self.noreport, self.exitcodes = (False, True)
        self.verbose_mails, self._verbose = (False, False)
        self._set_loglevel(logging.INFO)
        # Processing options:
        self.dryrun, self.maxsize, self.teachonly = (False, 120000, False)
        self.spamc, self.gmail = (False, False)
        # spamassassin options:
        self.movehamto, self.delete = (None, False)
        self.deletehigherthan, self.flag, self.expunge = (None, False, False)
        # Learning options:
        self.learnflagged, self.learnunflagged = (False, False)
        self.learnthendestroy, self.learnthenflag = (False, False)
        # Lockfile options:
        self.ignorelockfile = False
        self.lockfilename = os.path.join(xdg_cache_home, "isbg", "lock")
        self.lockfilegrace = 240.0
        # Password options (a vague level of obfuscation):
        self.passwdfilename, self.savepw = (None, False)
        # Trackfile options:
        self.trackfile, self.partialrun = (None, 50)

        try:
            self.interactive = sys.stdin.isatty()
        except AttributeError:
            self.logger.warning("Can't get info about if stdin is a tty")
            self.interactive = False

        # what we use to set flags on the original spam in imapbox
        self.spamflagscmd = "+FLAGS.SILENT"
        # and the flags we set them to (none by default)
        self.spamflags = []

    @staticmethod
    def set_filename(imapsets, filetype):
        """Set the filename of cached created files.

        If `filetype` is password, the file name start with `.isbg-`, else
        it starts with the filetype. A hash from the imapsets it's appended
        to it. The path of the file will be ``xdg_cache_home``/isbg/

        Args:
            imapsets (isbg.imaputils.ImapSettings): Imap setings instance.
            filetype (str): The file type.
        Returns:
            str: The full file path.

        """
        if filetype == "password":
            filename = os.path.join(xdg_cache_home, "isbg", ".isbg-")
        else:
            filename = os.path.join(xdg_cache_home, "isbg", filetype)
        return filename + imapsets.hash.hexdigest()

    @property
    def verbose(self):
        """Get the verbose property.

        :getter: Gets the verbose property.
        :setter: Sets verbose property.
        :type: bool
        """
        return self._verbose

    @verbose.setter
    def verbose(self, newval):
        """Set the verbose property and the required log level."""
        self._verbose = newval
        if self._verbose:
            self._set_loglevel(logging.DEBUG)
        else:
            self._set_loglevel(logging.INFO)

    def _set_loglevel(self, level):
        """Set the log level."""
        self.logger.setLevel(level)
        for handler in self.logger.handlers:
            handler.setLevel(level)

    def removelock(self):
        """Remove the lockfile."""
        if os.path.exists(self.lockfilename):
            os.remove(self.lockfilename)

    def assertok(self, res, *args):
        """Check that the return code is OK.

        It also prints out what happened (which would end
        up /dev/null'ed in non-verbose mode)
        """
        if 'uid FETCH' in args[0] and not self.verbose_mails:
            res = utils.shorten(res, 140)
        if 'SEARCH' in args[0]:
            res = utils.shorten(res, 140)
        self.logger.debug("{} = {}".format(args, res))
        if res[0] not in ["OK", "BYE"]:
            self.logger.error(
                __("{} returned {} - aborting".format(args, res)))
            raise ISBGError(__exitcodes__['imap'] if self.exitcodes else -1,
                            "\n%s returned %s - aborting\n" % (repr(args), res)
                            )

    def pastuid_read(self, uidvalidity, folder='inbox'):
        """Read the uids stored in a file for  a folder.

        pastuids_read keeps track of which uids we have already seen, so
        that we don't analyze them multiple times. We store its
        contents between sessions by saving into a file as Python
        code (makes loading it here real easy since we just source
        the file)
        """
        if self.trackfile is None:
            self.trackfile = ISBG.set_filename(self.imapsets, "track")
        pastuids = []
        try:
            with open(self.trackfile + folder, 'r') as rfile:
                struct = json.load(rfile)
                if struct['uidvalidity'] == uidvalidity:
                    pastuids = struct['uids']
        except Exception:  # pylint: disable=broad-except
            pass
        return pastuids

    def pastuid_write(self, uidvalidity, origpastuids, newpastuids,
                      folder='inbox'):
        """Write the uids in a file for the folder."""
        if self.dryrun:
            self.logger.debug(__(('Skipping pastuids write due to dryrun')))
            return

        if self.trackfile is None:
            self.trackfile = ISBG.set_filename(self.imapsets, "track")

        wfile = open(self.trackfile + folder, "w+")
        try:
            os.chmod(self.trackfile + folder, 0o600)
        except Exception:  # pylint: disable=broad-except
            pass
        self.logger.debug(__(('Writing pastuids for folder {}: {} ' +
                              'origpastuids, newpastuids: {}').format(
            folder, len(origpastuids), newpastuids)))
        struct = {
            'uidvalidity': uidvalidity,
            'uids': list(set(newpastuids + origpastuids))
        }
        json.dump(struct, wfile)
        wfile.close()

    def _do_lockfile_or_raise(self):
        """Create the lockfile or raise a error if it exists."""
        if (os.path.exists(self.lockfilename) and
                (os.path.getmtime(self.lockfilename) +
                    (self.lockfilegrace * 60) > time.time())):
            raise ISBGError(__exitcodes__['locked'],
                            "Lock file is present. Guessing isbg is " +
                            "already running. Exit.")
        else:
            lockfile = open(self.lockfilename, 'w')
            lockfile.write(repr(os.getpid()))
            lockfile.close()

            # Make sure to delete lock file
            atexit.register(self.removelock)

    def _do_get_password(self):
        """Get the password from file or prompt for it."""
        if (self.savepw is False and
                os.path.exists(self.passwdfilename) is True):
            try:
                sec = secrets.SecretIsbg(filename=self.passwdfilename,
                                         imapset=self.imapsets)
                self.imapsets.passwd = sec.get("password")
                del sec
                self.logger.debug("Successfully read password file")
            except Exception:  # pylint: disable=broad-except
                self.logger.exception('Error reading pw!')

        # do we have to prompt?
        if self.imapsets.passwd is None:
            if not self.interactive:
                raise ISBGError(__exitcodes__['ok'],
                                "You need to specify your imap password " +
                                "and save it with the --savepw switch")
            self.imapsets.passwd = getpass.getpass(
                "IMAP password for %s@%s: " % (
                    self.imapsets.user, self.imapsets.host))

    def _do_save_password(self):
        """Save password to the password file."""
        try:
            sec = secrets.SecretIsbg(filename=self.passwdfilename,
                                     imapset=self.imapsets)
            sec.set("password", self.imapsets.passwd)
            del sec
            self.logger.info('Note that passwords are saved in plain-text.')
        except Exception:  # pylint: disable=broad-except
            self.logger.exception('Error saving pw!')

    def do_list_imap(self):
        """List the imap boxes."""
        imap_list = self.imap.list()
        for x in imap_list[1]:
            x = re.sub(r'\(.*" (?=[a-zA-Z0-9])', "", x) # string formatting with
            self.logger.info(x)                         # lookbehind regex

    def do_spamassassin(self):
        """Do the spamassassin procesing.

        It creates a instance of :py:class:`~isbg.spamproc.SpamAssassin`
        every time that this method is called. The ``SpamAssasssin`` object
        would contact to the IMAP server to get the emails and to
        ``SpamAssassin`` command line to process them.

        """
        sa = spamproc.SpamAssassin.create_from_isbg(self)
        proc = None

        # SpamAssassin training: Learn spam
        s_learned = spamproc.Sa_Learn()
        if self.imapsets.learnspambox:
            uidvalidity = self.imap.get_uidvalidity(self.imapsets.learnspambox)
            origpastuids = self.pastuid_read(uidvalidity, 'spam')
            s_learned = sa.learn(self.imapsets.learnspambox, 'spam', None,
                                 origpastuids)
            self.pastuid_write(uidvalidity, s_learned.newpastuids,
                               s_learned.uids, 'spam')

        # SpamAssassin training: Learn ham
        h_learned = spamproc.Sa_Learn()
        if self.imapsets.learnhambox:
            uidvalidity = self.imap.get_uidvalidity(self.imapsets.learnhambox)
            origpastuids = self.pastuid_read(uidvalidity, 'ham')
            h_learned = sa.learn(self.imapsets.learnhambox, 'ham',
                                 self.movehamto, origpastuids)
            self.pastuid_write(uidvalidity, h_learned.newpastuids,
                               h_learned.uids, 'ham')

        if not self.teachonly:
            # check spaminbox exists by examining it
            self.imap.select(self.imapsets.spaminbox, 1)

            uidvalidity = self.imap.get_uidvalidity(self.imapsets.inbox)
            origpastuids = self.pastuid_read(uidvalidity)
            proc = sa.process_inbox(origpastuids)
            self.pastuid_write(uidvalidity, proc.newpastuids, proc.uids)

        if self.nostats is False:
            if self.imapsets.learnspambox is not None:
                self.logger.info(__(
                    "{}/{} spams learned".format(s_learned.learned,
                                                 s_learned.tolearn)))
            if self.imapsets.learnhambox:
                self.logger.info(__(
                    "{}/{} hams learned".format(h_learned.learned,
                                                h_learned.tolearn)))
            if not self.teachonly:
                self.logger.info(__(
                    "{} spams found in {} messages".format(proc.numspam,
                                                           proc.nummsg)))
                self.logger.info(__("{}/{} was automatically deleted".format(
                    proc.spamdeleted, proc.numspam)))

        return proc

    def do_imap_login(self):
        """Login to the imap."""
        self.imap = imaputils.login_imap(self.imapsets,
                                         logger=self.logger,
                                         assertok=self.assertok)

    def do_imap_logout(self):
        """Sign off from the imap connection."""
        self.imap.logout()

    def do_isbg(self):
        """Execute the main isbg process.

        It should be called to process the IMAP account. It returns a
        exitcode if its called from the command line and have the --exitcodes
        param.
        """
        if self.delete and not self.gmail and \
                "\\Deleted" not in self.spamflags:
            self.spamflags.append("\\Deleted")

        if self.trackfile is None:
            self.trackfile = ISBG.set_filename(self.imapsets, "track")

        if self.passwdfilename is None:
            self.passwdfilename = ISBG.set_filename(self.imapsets, "password")

        self.logger.debug(__("Lock file is {}".format(self.lockfilename)))
        self.logger.debug(__("Trackfile starts with {}".format(self.trackfile))
                          )
        self.logger.debug(__(
            "Password file is {}".format(self.passwdfilename)))
        self.logger.debug(__("SpamFlags are {}".format(self.spamflags)))

        # Acquire lockfilename or exit
        if self.ignorelockfile:
            self.logger.debug("Lock file is ignored. Continue.")
        else:
            self._do_lockfile_or_raise()

        # Figure out the password
        if self.imapsets.passwd is None:
            self._do_get_password()

        # ***** Main code starts here *****

        # Connection with the imaplib server
        self.do_imap_login()

        # Should we save it?
        if self.savepw:
            self._do_save_password()

        if self.imaplist:
            # List imap directories
            self.do_list_imap()
        else:
            # Spamassasin training and processing:
            proc = self.do_spamassassin()

        # sign off
        self.do_imap_logout()

        if self.exitcodes and __name__ == '__main__':
            if not self.teachonly:
                if proc.numspam == 0:
                    return __exitcodes__['newmsgs']
                if proc.numspam == proc.nummsg:
                    return __exitcodes__['newspam']
                return __exitcodes__['newmsgspam']

            return __exitcodes__['ok']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  utils.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

# From https://docs.python.org/3/howto/logging-cookbook.html
# Get free of the pylint logging-format-interpolation warning using __

"""Utils for isbg - IMAP Spam Begone."""

import os
import re
from subprocess import Popen, PIPE   # To call Popen

try:
    # C implementation:
    import cchardet
except ImportError:
    try:
        # Pure python implementation:
        import chardet
    except ImportError:
        pass


def detect_enc(byte_sring):
    """Try to detect a encoding.

    It uses the ``cchardet`` and ``chardet`` modules to detect the encoding.
    If none of them are installed, it returns None.

    Args:
        byte_string (str | bytes): the byte string to detect.

    Return:
        dict: with at least the 'encoding' informed as returned by
            :py:func:`cchardet.detect` and :py:func:`chardet.detect`.

    """
    try:
        ret = cchardet.detect(byte_sring)
    except NameError:
        ret = None

    if not ret or 'encoding' not in ret or not ret['encoding'] or \
            ret['encoding'] == '':
        try:
            ret = chardet.detect(byte_sring)
        except NameError:
            pass

    if not ret or not ret['encoding']:
        return {'encoding': None}

    return ret


def hexdigit(char):
    """Tanslate a hexadecimal character his decimal (int) value.

    Args:
        char (str): A hexadecimal number in base 16.
    Returns:
        int: the base 10 representation of the number.

    Raises:
        ValueError: if `char` is not a valid hexadecimal character.

    """
    if char >= '0' and char <= '9':
        return ord(char) - ord('0')
    if char >= 'a' and char <= 'f':
        return 10 + ord(char) - ord('a')
    if char >= 'A' and char <= 'F':
        return 10 + ord(char) - ord('A')
    raise ValueError(repr(char) + " is not a valid hexadecimal digit")


def hexof(string):
    """Translate a string to a string with its hexadecimal value.

    Args:
        string (str): A string to be translated.
    Returns:
        str: The translated string.

    Examples:
        >>> isbg.utils.hexof('isbg')
        '69736267'

    """
    res = ""
    for i in string:
        res = res + ("%02x" % ord(i))
    return res


def dehexof(string):
    """Tanslate a hexadecimal string to his string value.

    Args:
        string (str): A string containing a hexadecimal.
    Returns:
        str: The translated string.

    """
    res = ""
    while string:
        res = res + chr(16 * hexdigit(string[0]) + hexdigit(string[1]))
        string = string[2:]
    return res


def get_ascii_or_value(value):
    """Try to convert the contents of value to ascii string.

    When the `value` cannot be converted to an ascii string, it returns the
    value.

    Args:
        value (dict, list, str): The value to convert.
    Returns:
        The `value` object with its contents translated if it was possible.

    Note:
        We get the ``uids`` info as binary when using the methods of
        :py:class:`isbg.imaputils.IsbgImap4`.

    Examples:
        >>> get_ascii_or_value('isbg - IMAP Spam Begone')
        'isbg - IMAP Spam Begone'
        >>> d = {'isbg': (u'IMAP', [b'Spam', r'Begone'])}
        >>> get_ascii_or_value(d)
        {'isbg': ('IMAP', ['Spam', 'Begone'])}

    """
    def _get_ascii_or_value(val):
        """Try to convert to string.

        Args:
            val(str, byte): the value to convert.
        Returns:
            The value converted (or nor).

        """
        try:
            return val.decode('ascii')
        except UnicodeDecodeError:
            return val

    if isinstance(value, bytes):
        return _get_ascii_or_value(value)

    if isinstance(value, (list, tuple)):
        lis = []
        for v in value:
            lis.append(get_ascii_or_value(v))
        if isinstance(value, tuple):
            lis = tuple(lis)
        return lis

    if isinstance(value, dict):
        dic = {}
        for k, v in value.items():
            dic[get_ascii_or_value(k)] = get_ascii_or_value(v)
        return dic

    return value


def popen(cmd):
    """Create a :py:class:`subprocess.Popen` instance.

    It calls `Popen(cmd, stdin=PIPE, stdout=PIPE, close_fds=True)`.

    Args:
        cmd (str): The command to use in the call to Popen.
    Returns:
        subprocess.Popen: The `Popen` object.

    """
    if os.name == 'nt':
        return Popen(cmd, stdin=PIPE, stdout=PIPE)
    return Popen(cmd, stdin=PIPE, stdout=PIPE, close_fds=True)


def score_from_mail(mail):
    """
    Search the spam score from a mail as a string.

    The returning format is ``d.d/d.d<br>`` and it contains the score found
    in the email.

    Args:
        mail (str): A email.message.Message decoded.
    Returns:
        str: The score found in the mail message.

    """
    res = re.search(r"score=(-?\d+(?:\.\d+)?) required=(\d+(?:\.\d+)?)", mail)
    score = res.group(1) + "/" + res.group(2) + "\n"
    return score


def shorten(inp, length):
    """Short a dict or a list a tuple or a string to a maximus length.

    Args:
        inp (dict, list, tuple, str): The object to short.
        length (int): The length.
    Returns:
        the shorted object.

    """
    if isinstance(inp, dict):
        return dict([(k, shorten(v, length)) for k, v in inp.items()])
    elif isinstance(inp, (list, tuple)):
        lis = [shorten(x, length) for x in inp]
        if isinstance(inp, tuple):
            lis = tuple(lis)
        return lis
    return truncate(inp, length)


def truncate(inp, length):
    u"""Truncate a string to a maximum length.

    Args:
        inp (str): The string to be shortened to his maximum length.
        length (int): The length.

    Returns:
        (str): the shorted string.

        It adds ``…`` at the end if it is shortened.

    Raises:
        ValueError: If length is low than 1.

    """
    if length < 1:
        raise ValueError("length should be 1 or greater")
    if len(inp) > length:
        return repr(inp)[:length - 1] + '…'
    return inp


class BraceMessage(object):
    """Comodity class to format a string.

    You can call it using: py: class: `~__`

    Example:
        >> > from isbg.utils import __
        >> > __("ffoo, boo {}".format(a))

    """

    def __init__(self, fmt, *args, **kwargs):
        """Initialize the object."""
        self.fmt = fmt        #: The string to be formated.
        self.args = args      #: The `*args`
        self.kwargs = kwargs  #: The `**kwargs**`

    def __str__(self):
        """Return the string formated."""
        return self.fmt.format(*self.args, **self.kwargs)

    def __repr__(self):
        """Return the representation formated."""
        return self.__str__()


__ = BraceMessage  # pylint: disable=invalid-name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  test_utils.py
#  This file is part of utils.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""Test cases for utils module."""

import os
import sys

try:
    import pytest
except ImportError:
    pass

# We add the upper dir to the path
sys.path.insert(0, os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..')))
from isbg import utils  # noqa: E402


def test_detect_enc():
    """Test the detenct_enc function."""
    assert 'encoding' in utils.detect_enc(b"foo")
    assert 'encoding' in utils.detect_enc(b"")


def test_hexof_dehexof():
    """Test the dehexof function."""
    dehex = utils.dehexof("50402A")
    assert dehex == "P@*"
    assert utils.hexof(dehex) == "50402a"
    with pytest.raises(ValueError,
                       match=repr("G") + " is not a valid hexadecimal digit"):
        utils.dehexof("G")
        pytest.fail("Not error or unexpected error message")
    dehex = utils.dehexof("50402a")
    assert dehex == "P@*"
    assert utils.hexof(dehex) == "50402a"


def test_get_ascii_or_value():
    """Test get_ascii_or_value."""
    ret = utils.get_ascii_or_value(b'IMAP Spam Begone')
    assert ret == u'IMAP Spam Begone', 'should return IMAP Spam Begone'
    ret = utils.get_ascii_or_value(u'IMAP Spam Begone')
    assert ret == u'IMAP Spam Begone', 'should return IMAP Spam Begone'
    ret = utils.get_ascii_or_value('IMAP Spam Begone')
    assert ret == u'IMAP Spam Begone', 'should return IMAP Spam Begone'
    ret = utils.get_ascii_or_value(22)
    assert ret == 22, 'should return 22 (int)'

    d = {'isbg': (u'IMAP', [b'Spam', r'Begone'])}
    ret = utils.get_ascii_or_value(d)
    assert ret == {u'isbg': (u'IMAP', [u'Spam', u'Begone'])}, 'error'


def test_score_from_mail():
    """Test score_from_mail."""
    # Without score:
    fmail = open('tests/examples/spam.eml', 'rb')
    ftext = fmail.read()
    fmail.close()
    with pytest.raises(AttributeError):
        ret = utils.score_from_mail(ftext.decode(errors='ignore'))
        pytest.fail("Should rise AttributeError.")
    # With score:
    fmail = open('tests/examples/spam.from.spamassassin.eml', 'rb')
    ftext = fmail.read()
    fmail.close()
    ret = utils.score_from_mail(ftext.decode(errors='ignore'))
    assert ret == u"6.4/5.0\n", "Unexpected score."


def test_shorten():
    """Test the shorten function."""
    # We try with dicts:
    dic = {1: 'Option 1', 2: u'Option 2', 3: b'Option 3'}
    assert dic == utils.shorten(dic, 8), "The dicts should be the same."
    dic2 = utils.shorten(dic, 7)
    assert dic != dic2, "The dicts should be diferents."
    assert dic2[1] in ['u\'Opti…', '\'Optio…'], "Unexpected shortened string."
    assert dic2[2] in ['u\'Opti…', '\'Optio…'], "Unexpected shortened string."
    assert dic2[3] in ['\'Optio…', 'b\'Opti…'], "Unexpected shortened string."

    # We try with lists:
    ls = ['Option 1', 'Option 2', 'Option 3']
    assert ls == utils.shorten(ls, 8)
    ls2 = utils.shorten(ls, 7)
    for k in ls2:
        assert k in ['u\'Opti…', '\'Optio…'], "Unexpected shortened string."

    # We try with strings:
    assert "Option 1" == utils.shorten("Option 1", 8), \
        "Strings should be the same."
    assert utils.shorten("Option 1", 7) in ['u\'Opti…', "\'Optio…"], \
        "Strings should be diferents."

    # Others:
    with pytest.raises(TypeError):
        utils.shorten(None, 8)
        pytest.fail("None should raise a TypeError.")
    with pytest.raises(TypeError):
        utils.shorten(None, 7)
        pytest.fail("None should raise a TypeError.")
    with pytest.raises(TypeError):
        utils.shorten(False, 8)
        pytest.fail("None should raise a TypeError.")
    with pytest.raises(TypeError):
        utils.shorten(True, 7)
        pytest.fail("None should raise a TypeError.")
    with pytest.raises(TypeError):
        utils.shorten(1, 7)
        pytest.fail("int should raise a TypeError.")
    with pytest.raises(TypeError):
        utils.shorten(1.0, 7)
        pytest.fail("float should raise a TypeError.")
    with pytest.raises(ValueError):
        utils.shorten("123", 0)
        pytest.fail("length should be at least 1.")
    with pytest.raises(TypeError):
        assert utils.shorten([1, 2, 3], 2)
        pytest.fail("int should be not supported.")
    assert utils.shorten(["111", "2", "3"], 3) == ["111", "2", "3"]
    assert utils.shorten(("111", "2", "3"), 3) == ("111", "2", "3")


def test_BraceMessage():
    """Test BraceMessage."""
    ret = utils.BraceMessage("Test {} {}".format(1, "is one."))
    assert str(ret) == "Test 1 is one."
    ret = utils.__("Test {} {}".format(1, "is one."))
    assert str(ret) == "Test 1 is one."



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  test_secrets.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""Tests for secrets.py."""

import base64
import os
import sys

try:
    import pytest
except ImportError:
    pass

# We add the upper dir to the path
sys.path.insert(0, os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..')))
from isbg import secrets      # noqa: E402
from isbg import imaputils    # noqa: E402


class Test_SecretIsbg(object):
    """Test SecretIsbg class."""

    def test_get_and_set(self):
        """Test the get and set functions."""
        imapset = imaputils.ImapSettings()

        # Test with a erroneous filename.
        sec = secrets.SecretIsbg(filename="", imapset=imapset)
        assert sec.get("foo") is None
        with pytest.raises(EnvironmentError, match=r"\[Errno "):
            sec.set("foo", "boo")
            pytest.fail("A EnvironmentError should be raised.")

        # Test with a ok filename:
        sec = secrets.SecretIsbg(filename="tmp.txt", imapset=imapset)
        assert sec.get("foo") is None
        sec.set("foo1", "boo1")
        sec.set("foo2", "boo2")
        sec.set("foo3", "boo3")
        sec.set("foo3", "boo4")
        assert sec.get("foo3") == "boo4"

        with pytest.raises(ValueError, match="Key 'foo3' exists."):
            sec.set("foo3", "boo5", overwrite=False)
            pytest.fail("It should raise a ValueError")

        assert sec.get("foo3") == "boo4"

        # Remove the created keys:
        sec.delete("foo1")
        assert sec.get("foo1") is None
        sec.delete("foo2")
        assert sec.get("foo2") is None

        # Remove non existant key:
        with pytest.raises(ValueError, match="Key 'foo4' not"):
            sec.delete("foo4")
            pytest.fail("It should raise a ValueError")

        # Remove last key, it should delete the file:
        sec.delete("foo3")
        assert sec.get("foo3") is None
        with pytest.raises(EnvironmentError, match=r"\[Errno 2\]"):
            os.remove("tmp.txt")
            pytest.fail("A EnvironmentError should be raised.")

        # Remove non existant key in non existant file:
        with pytest.raises(ValueError, match="Key 'foo4' not"):
            sec.delete("foo4")
            pytest.fail("It should raise a ValueError")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
u"""isbg scans an IMAP Inbox and runs every entry against SpamAssassin.

For any entries that match, the message is copied to another folder,
and the original marked or deleted.

"""

from .isbg import ISBG, ISBGError, __version__, __exitcodes__, __license__
from .spamproc import learn_mail, test_mail
from .sa_unwrap import unwrap

__all__ = ["__version__", "__exitcodes__", "__license__", "learn_mail",
           "test_mail", "unwrap", "ISBG", "ISBGError"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  test_spamaproc.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""Tests for spamproc.py."""

import os
import sys
try:
    import pytest
except ImportError:
    pass

from email.errors import MessageError

# We add the upper dir to the path
sys.path.insert(0, os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..')))
from isbg import spamproc   # noqa: E402
from isbg import isbg       # noqa: E402
from isbg.imaputils import new_message  # noqa: E402


@pytest.mark.spamassassin
def test_learn_mail():
    """Tests for learn_mail, with a running SA instance."""
    fmail = open('tests/examples/spam.eml', 'rb')
    ftext = fmail.read()
    mail = new_message(ftext)
    fmail.close()

    # We forget the mail:
    spamproc.learn_mail(mail, 'forget')
    # We forget the mail:
    ret, ret_o = spamproc.learn_mail(mail, 'forget')
    assert ret is 6, "Mail should be already unlearned."
    # We try to learn it (as spam):
    ret, ret_o = spamproc.learn_mail(mail, 'spam')
    assert ret is 5, "Mail should have been learned"
    # The second time it should be already learned:
    ret, ret_o = spamproc.learn_mail(mail, 'spam')
    assert ret is 6, "Mail should be already learned."


@pytest.mark.spamassassin
def test_test_mail():
    """Tests for learn_mail, with a running SA instance."""
    fmail = open('tests/examples/spam.eml', 'rb')
    ftext = fmail.read()
    mail = new_message(ftext)
    fmail.close()

    # We test the mail with spamc:
    score1, code1, spamassassin_result = spamproc.test_mail(mail, True)
    score2, code2, spamassassin_result = spamproc.test_mail(mail, cmd=["spamc",
                                                 "-E", "--max-size=268435450"])
    assert score1 == score2, "The score should be the same."
    assert code1 == code2, "The return code should be the same."
    score, code, spamassassin_result = spamproc.test_mail("", True)
    assert score == u'-9999', 'It should return a error'
    assert code is None, 'It should return a error'

    # We test the mail with spamassassin:
    score3, code3, spamassassin_result = spamproc.test_mail(mail, False)
    score4, code4, spamassassin_result = spamproc.test_mail(mail, cmd=["spamassassin",
                                                  "--exit-code"])
    assert score3 == score4, "The score should be the same."
    assert code3 == code4, "The return code should be the same."
    score, code, spamassassin_result = spamproc.test_mail("", False)
    assert score == u'-9999', 'It should return a error'
    assert code is None, 'It should return a error'

    # We try a random cmds (existant and unexistant):
    score, code, spamassassin_result = spamproc.test_mail("", cmd=["echo"])
    assert score == u'-9999', 'It should return a error'
    assert code is None, 'It should return a error'
    with pytest.raises(OSError, match="No such file"):
        spamproc.test_mail(mail, cmd=["_____fooo___x_x"])
        pytest.fail("Should rise OSError.")


class Test_Sa_Learn(object):
    """Tests for SA_Learn."""

    def test_sa_learn(self):
        """Test for sa_learn."""
        learn = spamproc.Sa_Learn()
        assert learn.tolearn == 0
        assert learn.learned == 0
        assert len(learn.uids) == 0
        assert len(learn.newpastuids) == 0


class Test_Sa_Process(object):
    """Tests for SA_Process."""

    def test_sa_process(self):
        """Test for sa_process."""
        proc = spamproc.Sa_Process()
        assert proc.nummsg == 0
        assert proc.numspam == 0
        assert proc.spamdeleted == 0
        assert len(proc.uids) == 0
        assert len(proc.newpastuids) == 0


class Test_SpamAssassin(object):
    """Tests for SpamAssassin."""

    _kwargs = ['imap', 'spamc', 'logger', 'partialrun', 'dryrun',
               'learnthendestroy', 'gmail', 'learnthenflag', 'learnunflagged',
               'learnflagged', 'deletehigherthan', 'imapsets', 'maxsize',
               'noreport', 'spamflags', 'delete', 'expunge']

    def test__kwars(self):
        """Test _kwargs is up to date."""
        assert self._kwargs == spamproc.SpamAssassin()._kwargs

    def test___init__(self):
        """Test __init__."""
        sa = spamproc.SpamAssassin()
        # All args spected has been initialized:
        for k in self._kwargs:
            assert k in dir(sa)

        sa = spamproc.SpamAssassin(imap=0)
        for k in self._kwargs:
            assert k in dir(sa)

        with pytest.raises(TypeError, match="Unknown keyword"):
            sa = spamproc.SpamAssassin(imap2=0)
            pytest.fail("Should rise a error.")

    def test_cmd_save(self):
        """Test cmd_save."""
        sa = spamproc.SpamAssassin()
        assert sa.cmd_save == ['spamassassin']
        sa.spamc = True
        assert sa.cmd_save == ["spamc"]
        sa.spamc = False
        assert sa.cmd_save == ['spamassassin']

    def test_cmd_test(self):
        """Test cmd_test."""
        sa = spamproc.SpamAssassin()
        assert sa.cmd_test == ["spamassassin", "--exit-code"]
        sa.spamc = True
        assert sa.cmd_test == ["spamc", "-E", "--max-size=268435450"]
        sa.spamc = False
        assert sa.cmd_test == ["spamassassin", "--exit-code"]

    def test_create_from_isbg(self):
        """Test create_from_isbg."""
        sbg = isbg.ISBG()
        sa = spamproc.SpamAssassin.create_from_isbg(sbg)
        assert sa.imap is None  # pylint: disable=no-member
        assert sa.logger is not None

    def test_learn_checks(self):
        """Test learn checks."""
        sa = spamproc.SpamAssassin()
        with pytest.raises(isbg.ISBGError, match="Unknown learn_type"):
            sa.learn('Spam', '', None, [])
            pytest.fail("Should rise error.")

        with pytest.raises(isbg.ISBGError, match="Imap is required"):
            sa.learn('Spam', 'ham', None, [])
            pytest.fail("Should rise error.")

    def test_get_formated_uids(self):
        """Test get_formated_uids."""
        sbg = isbg.ISBG()
        sa = spamproc.SpamAssassin.create_from_isbg(sbg)

        # Test sorted and partialrun
        ret, oripast = sa.get_formated_uids(uids=[u'1 2 3'],
                                            origpastuids=[], partialrun=None)
        assert ret == [u'3', u'2', u'1']
        assert oripast == [], "List should be empty."

        ret, oripast = sa.get_formated_uids(uids=[u'1 2 3'],
                                            origpastuids=[], partialrun=3)
        assert ret == [u'3', u'2', u'1']
        assert oripast == [], "List should be empty."

        ret, oripast = sa.get_formated_uids(uids=[u'1 2 3'],
                                            origpastuids=[], partialrun=2)
        assert ret == [u'3', u'2']
        assert oripast == [], "List should be empty."

        # Test sorted and origpastuids. The uid '6' is not in the current uids,
        # and should be removed from the new origpastuids. And '3' should be
        # removed from the uids (it has been processed in the past).
        ret, oripast = sa.get_formated_uids(
            uids=[u'1 2 4 3'], origpastuids=[3, 1, 6], partialrun=2)
        print(oripast)
        print(ret)
        assert ret == [u'4', u'2']
        assert oripast == [3, 1], "Unexpected new orig past uids."

    def test_process_spam(self):
        """Test _process_spam."""
        sbg = isbg.ISBG()
        sa = spamproc.SpamAssassin.create_from_isbg(sbg)

        # OSError required when it's run without spamassassin (travis-cl)
        with pytest.raises((AttributeError, OSError, MessageError)):
            sa._process_spam(1, u"3/10\n", "", [], 0, "")
            pytest.fail("Should rise error, IMAP not created.")

        sa.noreport = True
        sa.deletehigherthan = 2
        sa._process_spam(1, u"3/10\n", "", [], 0, "")

    def test_process_inbox(self):
        """Test process_inbox."""
        sbg = isbg.ISBG()
        sa = spamproc.SpamAssassin.create_from_isbg(sbg)
        with pytest.raises(AttributeError, match="has no attribute"):
            sa.process_inbox([])
            pytest.fail("Should rise error, IMAP not created.")



#!/usr/bin/python
# -*- coding: utf-8 -*-

r"""
Parse an *rfc2822* email message and unwrap it if contains spam attached.

To know it it checks for an x-spam-type=original payload.

Does not perfectly preserve whitespace (esp. \r\n vs. \n and \t vs space),
but this should not impact spam-learning purposes.

Examples:
    It will return the original mail into a spamassassin mail:
        >>> import isbg.sa_unwrap
        >>> f = open('examples/spam.from.spamassassin.eml','rb')
        >>> spams = isbg.sa_unwrap.unwrap(f)
        >>> f.close()
        >>> for spam in spams:
        >>>     print(spam)

    or::

        $ isbg_sa_unwrap.py < examples/spam.from.spamassassin.eml
        $ isbg_sa_unwrap.py < examples/spam.eml

"""

import email
import email.message
from io import IOBase
import os
import sys

if __package__ is None and not hasattr(sys, 'frozen'):
    # direct call of __sa_unwrap__.py
    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))
import isbg  # noqa: E402

try:
    # Creating command-line interface
    from docopt import docopt, DocoptExit, printable_usage
except ImportError:
    sys.stderr.write("Missing dependency: docopt\n")
    raise

FILE_TYPES = (IOBase,)
PARSE_FILE = email.message_from_binary_file
MESSAGE = email.message_from_bytes


def sa_unwrap_from_email(msg):
    """Unwrap a email from the spamassasin email.

    Args:
        msg (email.message.Message): email to unwrap.

    Returns:
        [email.message.Message]: A list with the unwraped mails.

    """
    if msg.is_multipart():
        parts = []
        ploads = msg.get_payload()
        for pload in ploads:
            if pload.get_param('x-spam-type', '') == 'original':
                # We remove the headers added by spamassassin:
                if hasattr(pload, 'as_bytes'):
                    pl_bytes = pload.as_bytes()
                else:
                    pl_bytes = pload.as_string()
                el_idx = pl_bytes.index(b'\n\n')
                parts.append(MESSAGE(pl_bytes[el_idx + 2:]))
        if parts:  # len(parts) > 0
            return parts
    return None


def unwrap(mail):
    """Unwrap a email from the spamassasin email.

    the mail could be a email.message.Email, a file or a string or buffer.
    It ruturns a list with all the email.message.Email founds.

    Args:
        mail (email.message.Message, FILE_TYPES, str): the mail to unwrap.

    Returns:
        [email.message.Message]: A list with the unwraped mails.

    """
    if isinstance(mail, email.message.Message):
        return sa_unwrap_from_email(mail)
    if isinstance(mail, FILE_TYPES):  # files are also stdin...
        return sa_unwrap_from_email(PARSE_FILE(mail))
    mail = email.message_from_bytes(mail)
    return sa_unwrap_from_email(mail)


def __isbg_sa_unwrap_opts__():  # noqa: D207
    """isbg-sa-unwrap.py unwrap a mail bundled by SpamAssassin.

it parses a rfc2822 email message and unwrap it if contains spam attached.

Command line Options::

 Usage:
  isbg_sa_unwrap.py [--from <FROM_FILE>] [--to <TO_FILE>]
  isbg_sa_unwrap.py (-h | --help)
  isbg_sa_unwrap.py --usage
  isbg_sa_unwrap.py --version

 Options:
  -h, --help    Show the help screen.
  --usage       Show the usage information.
  --version     Show the version information.

  -f FILE, --from=FILE  Filename of the email to read and unwrap. If not
                        informed, the stdin will be used.
  -t FILE, --to=FILE    Filename to write the unwrapped  email. If not
                        informed, the stdout will be used.

"""


def isbg_sa_unwrap():
    """Run when this module is called from the command line."""
    try:
        opts = docopt(__isbg_sa_unwrap_opts__.__doc__,
                      version="isbg_sa_unwrap v" + isbg.__version__ +
                      ", from: " + os.path.abspath(__file__) + "\n\n" +
                      isbg.__license__)
    except DocoptExit:
        sys.stderr.write('Error with options!!!\n')
        raise

    if opts.get("--usage"):
        sys.stdout.write(
            "{}\n".format(printable_usage(__isbg_sa_unwrap_opts__.__doc__)))
        return

    if opts.get("--from"):
        file_in = open(opts["--from"], 'rb')
    else:
        file_in = sys.stdin

    if hasattr(file_in, 'buffer'):
        file_in = file_in.buffer    # pylint: disable=no-member

    spams = unwrap(file_in.read())
    file_in.close()

    if spams is None:
        sys.stderr.write("No spam into the mail detected.\n")
        return

    if opts.get("--to"):
        file_out = open(opts["--to"], 'wb')
    else:
        file_out = sys.stdout

    for spam in spams:
        file_out.write(spam.as_string())

    if file_out != sys.stdout:
        file_out.close()


if __name__ == '__main__':
    isbg_sa_unwrap()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Installation and build information for isbg."""

import ast
import os
import re
import sys

from setuptools import setup


# We get the isbg README.rst as package long description:
with open(os.path.join(os.path.dirname(__file__), 'README.rst'), 'rb') as f:
    LDESC = f.read().decode('utf-8')

# We get the version from isbg/isbg.py:
_VERSION_RE = re.compile(r'__version__\s+=\s+(.*)')
with open(os.path.join(os.path.dirname(__file__), 'isbg/isbg.py'), 'rb') as f:
    _VERSION = str(ast.literal_eval(_VERSION_RE.search(
        f.read().decode('utf-8')).group(1)))

# Only setup_require pytest-runner when test are called
needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

setup(
    name='isbg',
    version=_VERSION,  # to change it, change isbg/isbg.py: __version__
    description=(
        'a script that makes it easy to scan an IMAP inbox for spam using'
        + 'SpamAssassin and get your spam moved to another folder.'),
    long_description=LDESC,
    long_description_content_type="text/x-rst",
    keywords='email imap spamassasin filter',
    author='ISBG contributors (see CONTRIBUTORS file)',
    author_email='isbg@python.org',
    license='GPLv3',
    packages=['isbg'],
    entry_points={
        'console_scripts': [
            'isbg = isbg.__main__:main',
            'isbg_sa_unwrap = isbg.sa_unwrap:isbg_sa_unwrap',
        ]
    },
    install_requires=['docopt'],
    extras_require={
        'chardet': ['chardet'],
        'cchardet': ['cchardet'],
    },
    setup_requires=pytest_runner,
    tests_require=['pytest', 'mock', 'pytest-cov'],
    url='https://gitlab.com/isbg/isbg',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Email :: Post-Office :: IMAP',
        'Topic :: Communications :: Email :: Filters',
        'Topic :: Utilities',
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# test_sa_unwrap.py
# Copyright (C) 2017, Carles Muñoz Gorriz <carlesmu@internautas.org>
# This file is part of isbg.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""Test cases for sa_unwrap module."""

import os
import sys
import email.message

from unittest import mock

# We add the upper dir to the path
sys.path.insert(0, os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..')))
from isbg import sa_unwrap  # noqa: E402


def test_sa_unwrap_from_email():
    """Test function sa_unwrap_from_email."""
    """
    Test parsing a file with a message no multipart. It should return none.
    """
    f = open('tests/examples/spam.eml', 'rb')
    ret = sa_unwrap.sa_unwrap_from_email(sa_unwrap.PARSE_FILE(f))
    f.close()
    assert ret is None, "%r is not a None." % ret

    """
    Test creating a multipart message with 1 message from a file.
    It should return 1 Message.
    """
    f = open('tests/examples/spam.from.spamassassin.eml', 'rb')
    ftext = f.read()
    ret = sa_unwrap.sa_unwrap_from_email(sa_unwrap.MESSAGE(ftext))
    f.close()
    num_mails = 0
    for mail in ret:
        assert isinstance(mail, email.message.Message), \
            "%r is not a email.message.Message[]." % ret
        num_mails = num_mails + 1
    assert num_mails == 1, "%d mails found" % num_mails

    """
    Test a very simple message created from his bytes.
    It should return None.
    """
    # pylint: disable=no-member
    ret = sa_unwrap.sa_unwrap_from_email(email.message_from_bytes(b'0000'))
    assert ret is None, "%r is not a None." % ret

    """
    Test a very simple message created from his strings.
    It should return None.
    """
    ret = sa_unwrap.sa_unwrap_from_email(email.message_from_string('None'))
    assert ret is None, "%r is not a None." % ret


def test_unwrap():
    """Test funcition unwrap."""
    """Test unwrap of tests/examples/spam.from.spamassassin.eml --> 1 message."""
    f = open('tests/examples/spam.from.spamassassin.eml', 'rb')
    mails = sa_unwrap.unwrap(f)
    f.close()
    num_mails = 0
    for mail in mails:
        assert isinstance(mail, email.message.Message), \
            "%r is not a email.message.Message[]." % mails
        num_mails = num_mails + 1
    assert num_mails == 1, "%d mails found" % num_mails

    """Test unwrap of tests/examples/spam.eml --> None."""
    f = open('tests/examples/spam.eml', 'rb')
    mails = sa_unwrap.unwrap(f)
    f.close()
    assert mails is None

    """ Test a email.message.Message."""
    assert sa_unwrap.unwrap(email.message.Message()) is None


def test_isbg_sa_unwrap(capsys):
    """Test no multipart spam mail."""
    # Remove pytest options:
    args = sys.argv[:]
    del sys.argv[1:]

    sys.argv.append('-f')
    sys.argv.append('tests/examples/spam.eml')
    sa_unwrap.isbg_sa_unwrap()
    out, err = capsys.readouterr()
    assert err == "No spam into the mail detected.\n"

    # Restore pytest options:
    sys.argv = args[:]
    print(sys.argv)


def test_isbg_sa_unwrap_main(capsys):
    """Test multipart spam mail. Returns the multipart message."""
    # Remove pytest options:
    args = sys.argv[:]
    del sys.argv[1:]

    with mock.patch.object(sa_unwrap, "__name__", "__main__"):
        sys.argv.append('-f')
        sys.argv.append('tests/examples/spam.from.spamassassin.eml')
        sa_unwrap.isbg_sa_unwrap()
        out, err = capsys.readouterr()
        assert (out.startswith("Return-Path: <2587-84-162546-580") or
                out.startswith("Return-Path: \n <2587-84-162546-580"))

    # Restore pytest options:
    sys.argv = args[:]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  test_imaputils.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""Test cases for __main__ file."""

import os
import sys

try:
    import pytest
except ImportError:
    pass

from unittest import mock

# We add the upper dir to the path
sys.path.insert(0, os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..')))
from isbg import __main__, isbg  # noqa: E402


def test_parse_args():
    """Test parse_args."""
    # Remove pytest options:
    orig_args = sys.argv[:]

    # Parse command line (it should work):
    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--flag",
               "--noninteractive"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    __main__.parse_args(sbg)
    assert sbg.imapsets.host == "localhost"
    assert sbg.imapsets.user == "anonymous"
    assert sbg.imapsets.passwd == "none"

    # Parse with unknown option
    del sys.argv[1:]
    for op in ["--foo"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    with pytest.raises(isbg.ISBGError, match="[options]"):
        __main__.parse_args(sbg)
        pytest.fail("It should rise a docopt SystemExit")

    # Parse with bogus deletehigherthan
    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--deletehigherthan",
               "0"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    with pytest.raises(isbg.ISBGError, match="too small"):
        __main__.parse_args(sbg)
        pytest.fail("It should rise a too small ISBGError")

    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--deletehigherthan",
               "foo"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    with pytest.raises(isbg.ISBGError, match="Unrecognized score"):
        __main__.parse_args(sbg)
        pytest.fail("It should rise a unrecognized score" + "ISBGError")

    # Parse with ok deletehigherthan
    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--deletehigherthan",
               "8"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    __main__.parse_args(sbg)
    assert sbg.deletehigherthan == 8.0

    # Parse with bogus maxsize
    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--maxsize", "0"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    with pytest.raises(isbg.ISBGError, match="too small"):
        __main__.parse_args(sbg)
        pytest.fail("It should rise a too small ISBGError")

    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--maxsize", "foo"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    with pytest.raises(isbg.ISBGError, match="Unrecognised size"):
        __main__.parse_args(sbg)
        pytest.fail("It should rise a Unrecognised size" + "ISBGError")

    # Parse with ok maxsize
    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--maxsize", "12000"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    __main__.parse_args(sbg)
    assert sbg.maxsize == 12000

    # Parse with bogus partialrun
    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--partialrun", "-10"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    with pytest.raises(isbg.ISBGError, match="equal to 0 or higher"):
        __main__.parse_args(sbg)
        pytest.fail("It should rise a equal to 0 or higher " + "ISBGError")

    # Parse with 0 partialrun
    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--partialrun", "0"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    __main__.parse_args(sbg)
    assert sbg.partialrun is None

    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--dryrun", "--partialrun", "foo"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    with pytest.raises(isbg.ISBGError, match="must be a integer"):
        __main__.parse_args(sbg)
        pytest.fail("It should rise a invalid literal " + "ValueError")

    # Parse with ok partialrun and verbose and nossl
    del sys.argv[1:]
    for op in ["--imaphost", "localhost", "--imapuser", "anonymous",
               "--imappasswd", "none", "--verbose", "--partialrun", "10",
               "--nossl"]:
        sys.argv.append(op)
    sbg = isbg.ISBG()
    __main__.parse_args(sbg)
    assert sbg.partialrun == 10

    # Restore pytest options:
    del sys.argv[1:]
    sys.argv = orig_args[:]


def test_main():
    """Test main()."""
    # Remove pytest options:
    args = sys.argv[:]
    del sys.argv[1:]
    sys.argv.append("--ignorelockfile")

    with mock.patch.object(isbg, "__name__", "__main__"):
        with pytest.raises(SystemExit, match="10"):
            __main__.main()
            pytest.fail("Not error or unexpected error")

    with pytest.raises(SystemExit, match="10"):
        with pytest.raises(isbg.ISBGError, match="10"):
            __main__.main()
            pytest.fail("Not error or unexpected error message")
        pytest.fail("Not error or unexpected error")

    sys.argv.append("--version")
    with mock.patch.object(isbg, "__name__", "__main__"):
        with pytest.raises(SystemExit):
            __main__.main()
            pytest.fail("Not error or unexpected error")

    sys.argv.append("--exitcodes")
    with mock.patch.object(isbg, "__name__", "__main__"):
        with pytest.raises(SystemExit):
            __main__.main()
            pytest.fail("Not error or unexpected error")

    del sys.argv[1:]
    sys.argv.append("--usage")
    with mock.patch.object(isbg, "__name__", "__main__"):
        with pytest.raises(SystemExit):
            __main__.main()
            pytest.fail("Not Exit or unexpected error")

    # Restore pytest options:
    sys.argv = args[:]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  spamaproc.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""Spam processing module for isbg."""
# pylint: disable=no-member

import isbg

from isbg import imaputils
from isbg import sa_unwrap
from isbg import utils

from .utils import __

import logging

#: Used to detect already our successfully (un)learned messages.
__spamc_msg__ = {
    'already': 'Message was already un/learned',
    'success': 'Message successfully un/learned'
}


def learn_mail(mail, learn_type):
    """Process a email and try to learn or unlearn it.

    Args:
        mail (email.message.Message): email to learn.
        learn_type (str): ```spam``` to learn spam, ```ham``` to learn
            nonspam or ```forget```.
    Returns:
        int, int: It returns a pair of `int`

        The first integer:
            A return code of ``6`` means it was already learned or forgotten,
            a return code of ``5`` means it has been learned or forgotten,
            a ``-9999`` means an error communicating with ``spamc``. If
            ``spamc`` returns an exit code, it returns it.

        The second integer:
            It's the original exit code from ``spamc``

    Notes:
        See `Exit Codes` section of the man page of ``spamc`` for more
        information about other exit codes.

    """
    out = ""
    orig_code = None
    proc = utils.popen(["spamc", "--learntype=" + learn_type])
    try:
        out = proc.communicate(imaputils.mail_content(mail))
        code = int(proc.returncode)
        orig_code = code
    except Exception:  # pylint: disable=broad-except
        code = -9999

    proc.stdin.close()

    if code == 0:
        out = out[0].decode(errors='ignore').strip()
        if out == __spamc_msg__['already']:
            code = 6
        elif out == __spamc_msg__['success']:
            code = 5

    return code, orig_code


def test_mail(mail, spamc=False, cmd=False):
    """Test a email with spamassassin."""
    score = "0/0\n"
    orig_code = None
    spamassassin_result = None
    returncode = None

    if cmd:
        satest = cmd
    elif spamc:
        # let spamc process mails larger than 500 KB if ISBG forwards them
        satest = ["spamc", "-E", "--max-size=268435450"]
    else:
        satest = ["spamassassin", "--exit-code"]

    proc = utils.popen(satest)

    try:
        spamassassin_result = proc.communicate(imaputils.mail_content(mail))[0]
        returncode = proc.returncode
        proc.stdin.close()
        score = utils.score_from_mail(spamassassin_result.decode(errors='ignore'))

    except Exception:  # pylint: disable=broad-except
        score = "-9999"

    return score, returncode, spamassassin_result


class Sa_Learn(object):
    """Commodity class to store information about learning processes."""

    def __init__(self):
        """Initialize `SA_Learn`."""
        self.tolearn = 0         #: Number of messages to learn.
        self.learned = 0         #: Number of messages learned.
        self.uids = []           #: The list of ``uids``.
        self.newpastuids = []    #: The new past ``uids``.


class Sa_Process(object):
    """Commodity class to store information about processes."""

    def __init__(self):
        """Initialize `SA_Process`."""
        self.nummsg = 0          #: Number of processed messages.
        self.numspam = 0         #: Number of spams found.
        self.spamdeleted = 0     #: Number of deleted spam.
        self.uids = []           #: The list of ``uids``.
        self.newpastuids = []    #: The new past ``uids``.


class SpamAssassin(object):
    """Learn and process spams from a imap account.

    You usually will create an instance of it using
    :py:func:`create_from_isbg`:

        >>> sa = isbg.spamproc.SpamAssassin.create_from_isbg(self)

    Or, if you are extending :py:class:`~isbg.ISBG`, it is created every
    time that you call to :py:func:`isbg.ISBG.do_spamassassin`.

    """

    #: key args required when initialized.
    _required_kwargs = []

    #: Key args that will be used.
    _kwargs = ['imap', 'spamc', 'logger', 'partialrun', 'dryrun',
               'learnthendestroy', 'gmail', 'learnthenflag', 'learnunflagged',
               'learnflagged', 'deletehigherthan', 'imapsets', 'maxsize',
               'noreport', 'spamflags', 'delete', 'expunge']

    def __init__(self, **kwargs):
        """Initialize a SpamAssassin object."""
        for k in self._required_kwargs:
            if k not in kwargs:
                raise TypeError("Missed required keyword argument: {}".format(
                                k))

        for k in self._kwargs:
            setattr(self, k, None)

        for k in kwargs:
            if k not in self._kwargs:
                raise TypeError("Unknown keyword argument: {}".format(k))
            setattr(self, k, kwargs[k])

        # pylint: disable=access-member-before-definition
        if self.logger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.addHandler(logging.StreamHandler())

        # what we use to set flags on the original spam in imapbox
        self.spamflagscmd = "+FLAGS.SILENT"

    @property
    def cmd_save(self):
        """Is the command that dumps out a munged message including report."""
        if self.spamc:  # pylint: disable=no-member
            return ["spamc"]
        return ["spamassassin"]

    @property
    def cmd_test(self):
        """Is the command to use to test if the message is spam."""
        if self.spamc:  # pylint: disable=no-member
            return ["spamc", "-E", "--max-size=268435450"]
        return ["spamassassin", "--exit-code"]

    @classmethod
    def create_from_isbg(cls, sbg):
        """Return a instance with the required args from ```ISBG```.

        Args:
            sbg (isbg.ISBG): His attributes will be used to initialize a
                SpamAssassin instance.
        Returns:
            SpamAssassin: A new created object with the required attributes
                based based `sbg` attributes.

        """
        kw = dict()
        for k in cls._kwargs:
            kw[k] = getattr(sbg, k)
        return SpamAssassin(**kw)

    @staticmethod
    def get_formated_uids(uids, origpastuids, partialrun):
        """Get the uids formated.

        Args:
            uids (list(str)): The new ``uids``. It's formated as:
                ```['1 2 3 4']```
            origpastuids (list(int)): The original past ``uids``.
            partialrun (int): If not none the number of ``uids`` to return.
        Returns:
            list(str): The ``uids`` formated.

            It sorts the uids, remove those that are in `origpastuids` and
            returns the number defined by `partialrun`. If `partialrun` is
            ```None``` it return all.

        """
        uids = sorted(uids[0].split(), key=int, reverse=True)
        newpastuids = [u for u in origpastuids if str(u) in uids]
        uids = [u for u in uids if int(u) not in newpastuids]
        # Take only X elements if partialrun is enabled
        if partialrun:
            uids = uids[:int(partialrun)]
        return uids, newpastuids

    def learn(self, folder, learn_type, move_to, origpastuids):
        """Learn the spams (and if requested deleted or move them).

        Args:
            folder (str): The IMAP folder.
            leart_type (str): ```spam``` to learn spam, ```ham``` to learn
                nonspam.
            move_to (str): If not ```None```, the imap folder where the emails
                will be moved.
            origpastuids (list(int)): ``uids`` to not process.
        Returns:
            Sa_Learn:
                It contains the information about the result of the process.

            It will call ``spamc`` to learn the emails.

        Raises:
            isbg.ISBGError: if learn_type is unknown.


        TODO:
            Add suport to ``learn_type=forget``.

        """
        sa_learning = Sa_Learn()

        # Sanity checks:
        if learn_type not in ['spam', 'ham']:
            raise isbg.ISBGError(-1, message="Unknown learn_type")
        if self.imap is None:
            raise isbg.ISBGError(-1, message="Imap is required")

        self.logger.debug(__(
            "Teach {} to SA from: {}".format(learn_type, folder)))

        self.imap.select(folder)
        if self.learnunflagged:
            _, uids = self.imap.uid("SEARCH", None, "UNFLAGGED")
        elif self.learnflagged:
            _, uids = self.imap.uid("SEARCH", None, "(FLAGGED)")
        else:
            _, uids = self.imap.uid("SEARCH", None, "ALL")

        uids, sa_learning.newpastuids = SpamAssassin.get_formated_uids(
            uids, origpastuids, self.partialrun)

        sa_learning.tolearn = len(uids)

        for uid in uids:
            mail = imaputils.get_message(self.imap, uid, logger=self.logger)

            # Unwrap spamassassin reports
            unwrapped = sa_unwrap.unwrap(mail)
            if unwrapped is not None:
                self.logger.debug(__("{} Unwrapped: {}".format(
                    uid, utils.shorten(imaputils.mail_content(
                        unwrapped[0]), 140))))

            if unwrapped is not None and unwrapped:  # len(unwrapped)>0
                mail = unwrapped[0]

            if self.dryrun:
                self.logger.warning("Skipped learning due to dryrun!")
                continue
            else:
                code, code_orig = learn_mail(mail, learn_type)

            if code == -9999:  # error processing email, try next.
                self.logger.exception(__(
                    'spamc error for mail {}'.format(uid)))
                self.logger.debug(repr(imaputils.mail_content(mail)))
                continue

            if code in [69, 74]:
                raise isbg.ISBGError(
                    isbg.__exitcodes__['flags'],
                    "spamassassin is misconfigured (use --allow-tell)")

            if code == 5:  # learned.
                sa_learning.learned += 1
                self.logger.debug(__(
                    "Learned {} (spamc return code {})".format(uid,
                                                               code_orig)))

            elif code == 6:  # already learned.
                self.logger.debug(__(
                    "Already learned {} (spamc return code {})".format(
                        uid, code_orig)))

            elif code == 98:  # too big.
                self.logger.warning(__(
                    "{} is too big (spamc return code {})".format(
                        uid, code_orig)))

            else:
                raise isbg.ISBGError(-1, ("{}: Unknown return code {} from " +
                                          "spamc").format(uid, code_orig))

            sa_learning.uids.append(int(uid))

            if not self.dryrun:
                if self.learnthendestroy:
                    if self.gmail:
                        self.imap.uid("COPY", uid, "[Gmail]/Trash")
                    else:
                        self.imap.uid("STORE", uid, self.spamflagscmd,
                                      "(\\Deleted)")
                elif move_to is not None:
                    self.imap.uid("COPY", uid, move_to)
                elif self.learnthenflag:
                    self.imap.uid("STORE", uid, self.spamflagscmd,
                                  "(\\Flagged)")

        return sa_learning

    def _process_spam(self, uid, score, mail, spamdeletelist, code, spamassassin_result):
        self.logger.debug(__("{} is spam".format(uid)))

        if (self.deletehigherthan is not None and
                float(score.split('/')[0]) > self.deletehigherthan):
            spamdeletelist.append(uid)
            return False

        # do we want to include the spam report
        if self.noreport is False:
            if self.dryrun:
                self.logger.info("Skipping report because of --dryrun")
            else:
                new_mail = spamassassin_result
                if new_mail == u"-9999":
                    self.logger.exception(
                        '{} error for mail {} (ret code {})'.format(
                            self.cmd_save, uid, code))
                    self.logger.debug(repr(imaputils.mail_content(mail)))
                    if uid in spamdeletelist:
                        spamdeletelist.remove(uid)
                    return False

                res = self.imap.append(self.imapsets.spaminbox, None, None,
                                       new_mail)
                # The above will fail on some IMAP servers for various
                # reasons. We print out what happened and continue
                # processing
                if res[0] != 'OK':
                    self.logger.error(__(
                        ("{} failed for uid {}: {}. Leaving original" +
                         "message alone.").format(
                            repr(["append", self.imapsets.spaminbox,
                                  "{email}"]),
                            repr(uid), repr(res))))
                    if uid in spamdeletelist:
                        spamdeletelist.remove(uid)
                    return False
        else:
            # No report:
            if self.dryrun:
                self.logger.info("Skipping copy to spambox because" +
                                 " of --dryrun")
            else:
                # just copy it as is
                self.imap.uid("COPY", uid, self.imapsets.spaminbox)

        return True

    def process_inbox(self, origpastuids):
        """Run spamassassin in the folder for spam."""
        sa_proc = Sa_Process()

        spamlist = []
        spamdeletelist = []

        # select inbox
        self.imap.select(self.imapsets.inbox, 1)

        # get the uids of all mails with a size less then the maxsize
        _, uids = self.imap.uid("SEARCH", None, "SMALLER", str(self.maxsize))

        uids, sa_proc.newpastuids = SpamAssassin.get_formated_uids(
            uids, origpastuids, self.partialrun)

        self.logger.debug(__('Got {} mails to check'.format(len(uids))))

        if self.dryrun:
            processednum = 0
            fakespammax = 1
            processmax = 5

        # Main loop that iterates over each new uid we haven't seen before
        for uid in uids:
            # Retrieve the entire message
            mail = imaputils.get_message(self.imap, uid, sa_proc.uids,
                                         logger=self.logger)

            # Unwrap spamassassin reports
            unwrapped = sa_unwrap.unwrap(mail)
            if unwrapped is not None and unwrapped:  # len(unwrapped) > 0
                mail = unwrapped[0]

            # Feed it to SpamAssassin in test mode
            if self.dryrun:
                if processednum > processmax:
                    break
                if processednum < fakespammax:
                    self.logger.info("Faking spam mail")
                    score = "10/10"
                    code = 1
                else:
                    self.logger.info("Faking ham mail")
                    score = "0/10"
                    code = 0
                processednum = processednum + 1
                spamassassin_result = None  # since dryrun doesn't run
                                            # test_mail()
            else:
                score, code, spamassassin_result = test_mail(mail, cmd=self.cmd_test)
                if score == "-9999":
                    self.logger.exception(__(
                        '{} error for mail {}'.format(self.cmd_test, uid)))
                    self.logger.debug(repr(mail))
                    uids.remove(uid)
                    continue

            if score == "0/0\n":
                raise isbg.ISBGError(isbg.__exitcodes__['spamc'],
                                     "spamc -> spamd error - aborting")

            self.logger.debug(__(
                "Score for uid {}: {}".format(uid, score.strip())))

            if code != 0:
                # Message is spam, delete it or move it to spaminbox
                # (optionally with report)
                if not self._process_spam(uid, score, mail, spamdeletelist, code, spamassassin_result):
                    continue
                spamlist.append(uid)

        sa_proc.nummsg = len(uids)
        sa_proc.spamdeleted = len(spamdeletelist)
        sa_proc.numspam = len(spamlist) + sa_proc.spamdeleted

        # If we found any spams, now go and mark the original messages
        if sa_proc.numspam or sa_proc.spamdeleted:
            if self.dryrun:
                self.logger.info('Skipping labelling/expunging of mails ' +
                                 ' because of --dryrun')
            else:
                self.imap.select(self.imapsets.inbox)
                # Only set message flags if there are any
                if self.spamflags:  # len(self.smpamflgs) > 0
                    for uid in spamlist:
                        self.imap.uid("STORE", uid, self.spamflagscmd,
                                      imaputils.imapflags(self.spamflags))
                        sa_proc.newpastuids.append(uid)
                # If its gmail, and --delete was passed, we actually copy!
                if self.delete and self.gmail:
                    for uid in spamlist:
                        self.imap.uid("COPY", uid, "[Gmail]/Trash")
                # Set deleted flag for spam with high score
                for uid in spamdeletelist:
                    if self.gmail is True:
                        self.imap.uid("COPY", uid, "[Gmail]/Trash")
                    else:
                        self.imap.uid("STORE", uid, self.spamflagscmd,
                                      "(\\Deleted)")
                if self.expunge:
                    self.imap.expunge()

        return sa_proc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  test_imaputils.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""Test cases for isbg module."""

import email
import logging
import os
import sys
try:
    import pytest
except ImportError:
    pass

from socket import gaierror

# We add the upper dir to the path
sys.path.insert(0, os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..')))
from isbg import imaputils  # noqa: E402


def test_mail_content():
    """Test mail_content function."""
    with pytest.raises(email.errors.MessageError):
        imaputils.mail_content(None)
        pytest.fail("mail 'None' is not a email.message.Message.")
    fmail = open('tests/examples/spam.from.spamassassin.eml', 'rb')
    ftext = fmail.read()
    mail = imaputils.new_message(ftext)
    fmail.close()
    assert isinstance(imaputils.mail_content(mail), (str, bytes))


def test_new_message():
    """Test new_message function."""
    fmail = open('tests/examples/spam.from.spamassassin.eml', 'rb')
    ftext = fmail.read()
    mail = imaputils.new_message(ftext)
    fmail.close()
    assert isinstance(mail, email.message.Message), \
        "%r is not a email.message.Message." % mail

    mail = imaputils.new_message("Foo")
    assert isinstance(mail, email.message.Message), \
        "%r is not a email.message.Message." % mail

    with pytest.raises(TypeError):
        imaputils.new_message(None)
    with pytest.raises(TypeError):
        imaputils.new_message(body="")
    with pytest.raises(TypeError, match="cannot be empty"):
        imaputils.new_message(body=b"")
    foo = imaputils.new_message(body=u"From:ñ@ñ.es")
    assert isinstance(foo, email.message.Message)


def test_get_message():
    """Test get_message."""
    # FIXME:
    pass


def test_imapflags():
    """Test imapflags."""
    assert imaputils.imapflags(['foo', 'boo']) == '(foo,boo)'


def test_login_imap():
    """Test login_imap."""
    with pytest.raises(TypeError, match="ImapSettings"):
        imaputils.login_imap(None)
        pytest.fail("spected a ImapSettings")

    imapsets = imaputils.ImapSettings()
    imapsets.host = ''  # don't try to connect to internet
    imapsets.nossl = True
    with pytest.raises(Exception, match="[Errno -5]"):
        imaputils.login_imap(imapsets, logger=logging.getLogger(__name__))
        pytest.fail("No address associated with hostname")
    # FIXME: require network


class TestImapSettings(object):
    """Test object ImapSettings."""

    def test(self):
        """Test the object."""
        imapset = imaputils.ImapSettings()
        imaphash = imapset.hash
        assert imapset.hash == imaphash
        assert imapset.hash.hexdigest() == '56fdd686137c8645d44024096a0ed441'
        assert imapset.hash.hexdigest() == '56fdd686137c8645d44024096a0ed441'
        assert imapset.hash.hexdigest() == '56fdd686137c8645d44024096a0ed441'
        imapset.host = '127.0.0.1'
        assert imapset.hash.hexdigest() == 'ca057ebec07690c05f64959fff011c8d'
        assert imapset.hash.hexdigest() == 'ca057ebec07690c05f64959fff011c8d'
        assert imapset.hash.hexdigest() == 'ca057ebec07690c05f64959fff011c8d'
        imapset.host = 'localhost'
        assert imapset.hash.hexdigest() == '56fdd686137c8645d44024096a0ed441'
        assert imapset.hash.hexdigest() == '56fdd686137c8645d44024096a0ed441'
        assert imapset.hash.hexdigest() == '56fdd686137c8645d44024096a0ed441'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  imaputils.py
#  This file is part of isbg.
#
#  Copyright 2018 Carles Muñoz Gorriz <carlesmu@internautas.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""Imap utils module for isbg - IMAP Spam Begone."""

import email          # To easily encapsulated emails messages
import email.message  # required for typing.TypeVar to work in py3
import imaplib
import re             # For regular expressions
import socket         # to catch the socket.error exception
import time

from hashlib import md5

from isbg import utils
from .utils import __

from typing import List, TypeVar, Union

Email = TypeVar(email.message.Message)
Uid = Union[int, str]
Uids = List[int]


def mail_content(mail):
    # type: (Email) -> AnyStr
    """Get the email message content.

    Args:
        mail (email.message.Message): The email message.

    Returns:
        :obj:`bytes` or :obj:`str`: The contents, with headers, of the email
        message. It returns `bytes`.

    Raises:
        email.errors.MessageError:  if mail is neither *bytes* nor *str*.

    """
    if not isinstance(mail, email.message.Message):
        raise email.errors.MessageError(
            "mail '{}' is not a email.message.Message.".format(repr(mail)))
    try:
        return mail.as_bytes()
    except UnicodeEncodeError:
        return mail.as_string()


def new_message(body):
    # type: (AnyStr) -> Email
    """Get a email.message from a body email.

    Note: If there are problems encoding it, it will replace it to *ascii*.

    Args:
        body (:obj:`bytes` or :obj:`str`): The content, with or without
            headers, of a email message.

    Returns:
        email.message.Message: The object representing it.

    Raises:
        TypeError: If the content is empty.

    """
    mail = None

    if isinstance(body, bytes):
        mail = email.message_from_bytes(body)  # pylint: disable=no-member
        if mail.as_bytes() in [b'', b'\n']:
            raise TypeError(
                __("body '{}' cannot be empty.".format(repr(body))))
        return mail

    try:
        mail = email.message_from_string(body)
    except UnicodeEncodeError:
        body = body.encode("ascii", errors='replace')
        mail = email.message_from_string(body)
    if mail.as_string() in ['', '\n']:
        raise TypeError(
            __("body '{}' cannot be empty.".format(repr(body))))
    return mail


def get_message(imap, uid, append_to=None, logger=None):
    # type: (IsbgImap4, Uid, Optional[Uids], Optional[logging.Logger]) -> Email
    """Get a message by *uid* and optionally append it to a list.

    Args:
        imap (IsbgImap4): The imap helper object with the connection.
        uid (:obj:`int` or :obj:`str`): A integer value with the *uid* of a
            message to fetch from the *imap* connection.
        append_to (:obj:`list` of :obj:`int`), optional): The integer value of
            *uid* is appended to this list. Defaults to *None*.
        logger (logging.Logger, optional): When a error is raised fetching the
            mail a warning is written to this logger. Defaults to *None*.

    Returns:
        email.message.Message: The message fetched from the *imap* connection.

    """
    res = imap.uid("FETCH", uid, "(BODY.PEEK[])")
    mail = email.message.Message()  # an empty email
    if res[0] != "OK":
        try:
            body = res[1][0][1]
            mail = new_message(body)
        except Exception:  # pylint: disable=broad-except
            logger.warning(__(
                ("Confused - rfc822 fetch gave {} - The message was " +
                 "probably deleted while we were running").format(res)))
    else:
        body = res[1][0][1]
        mail = new_message(body)

    if append_to is not None:
        append_to.append(int(uid))

    return mail


def imapflags(flaglist):
    # type: (List[str]) -> str
    """Transform a list to a string as expected for the IMAP4 standard.

    Example:
        >>> imapflags(['foo', 'boo')]
        '(foo,boo)'

    Args:
        flaglist (:obj:`list` of :obj:`str`): The flag list to transform.

    Returns:
        str: A string with the flag list.

    """
    return '(' + ','.join(flaglist) + ')'


def bytes_to_ascii(func):
    """Decorate a method to return his return value as *ascii*."""
    def func_wrapper(cls, *args, **kwargs):
        return utils.get_ascii_or_value(func(cls, *args, **kwargs))
    return func_wrapper


def assertok(name):
    """Decorate with *assertok*."""
    def assertok_decorator(func):
        def func_wrapper(cls, *args, **kwargs):
            res = func(cls, *args, **kwargs)
            if cls.assertok:
                if name == 'login':
                    cls.assertok(res, name, args[0], 'xxxxxxxx')
                elif name == 'uid':
                    cls.assertok(res, name + " " + args[0], args[1:])
                else:
                    cls.assertok(res, name, *args, **kwargs)
            return res
        return func_wrapper
    return assertok_decorator


class IsbgImap4(object):
    """Proxy class for :obj:`imaplib.IMAP4` and :obj:`imaplib.IMAP4_SSL`.

    It calls to the *IMAP4* or *IMAP4_SSL* methods but before it adds them
    decorators to log the calls and to try to convert the returns values to
    str.

    The only original method is ``get_uidvalidity``, used to return the current
    *uidvalidity* from a mailbox.

    """

    def __init__(self, host='', port=143, nossl=False, assertok=None):
        """Create a imaplib.IMAP4[_SSL] with an assertok method."""
        self.assertok = assertok
        self.nossl = nossl
        if nossl:
            self.imap = imaplib.IMAP4(host, port)
        else:
            self.imap = imaplib.IMAP4_SSL(host, port)

    # @assertok('append')  <-- it fails in some servers
    @bytes_to_ascii
    def append(self, mailbox, flags, date_time, message):
        """Append message to named mailbox."""
        return self.imap.append(mailbox, flags, date_time, message)

    @assertok('cabability')
    @bytes_to_ascii
    def capability(self):
        """Fetch capabilities list from server."""
        return self.imap.capability()

    @assertok('expunge')
    @bytes_to_ascii
    def expunge(self):
        """Permanently remove deleted items from selected mailbox."""
        return self.imap.expunge()

    @assertok('list')
    @bytes_to_ascii
    def list(self, directory='""', pattern='*'):
        """List mailbox names in directory matching pattern."""
        return self.imap.list(directory, pattern)

    @assertok('login')
    @bytes_to_ascii
    def login(self, user, passwd):
        """Identify client using plain text password."""
        return self.imap.login(user, passwd)

    @assertok('logout')
    @bytes_to_ascii
    def logout(self):
        """Shutdown connection to server."""
        return self.imap.logout()

    @assertok('status')
    @bytes_to_ascii
    def status(self, mailbox, names):
        """Request named status conditions for mailbox."""
        return self.imap.status(mailbox, names)

    @assertok('select')
    @bytes_to_ascii
    def select(self, mailbox='INBOX', readonly=False):
        """Select a Mailbox."""
        return self.imap.select(mailbox, readonly)

    @assertok('uid')
    @bytes_to_ascii
    def uid(self, command, *args):
        """Execute "command arg ..." with messages identified by UID."""
        return self.imap.uid(command, *args)

    def get_uidvalidity(self, mailbox):
        """Validate a mailbox.

        Args:
            mailbox (str): the mailbox to check for its *uidvalidity*.

        Returns:
            int: The *uidvalidity* returned from the *imap* server. If it
            cannot be decoded, it returns 0.

        """
        uidvalidity = 0
        mbstatus = self.imap.status(mailbox, '(UIDVALIDITY)')
        if mbstatus[0] == 'OK':
            body = mbstatus[1][0].decode()
            uidval = re.search('UIDVALIDITY ([0-9]+)', body)
            if uidval is not None:
                uidvalidity = int(uidval.groups()[0])
        return uidvalidity


def login_imap(imapsets, logger=None, assertok=None):
    """Login to the imap server."""
    if not isinstance(imapsets, ImapSettings):
        raise TypeError("imapsets is not a ImapSettings")

    max_retry = 10
    retry_time = 0.60   # seconds
    for retry in range(1, max_retry + 1):
        try:
            imap = IsbgImap4(imapsets.host, imapsets.port, imapsets.nossl,
                             assertok)
            break   # ok, exit from loop
        except socket.error as exc:
            if logger:
                logger.warning(__(
                    ("Error in IMAP connection: {} ... retry {} of {}"
                     ).format(exc, retry, max_retry)))
            if retry >= max_retry:
                raise Exception(exc)
            else:
                time.sleep(retry_time)
    if logger:
        logger.debug(__("Server capabilities: {}".format(
            imap.capability()[1])))
    if imapsets.nossl and logger:
        logger.warning("WARNING: Using insecure IMAP connection: without SSL.")
    # Authenticate (only simple supported)
    imap.login(imapsets.user, imapsets.passwd)
    return imap


class ImapSettings(object):
    """Class to store the *imap* and imap folders settings."""

    def __init__(self):
        """Set Imap settings."""
        self.host = 'localhost'      #: IMAP host name or IP.
        self.port = 143              #: IMAP port to connect.
        self.user = ''               #: IMAP user name.
        self.passwd = None           #: Password for the IMAP user name.
        self.nossl = False           #: Not use ssl for IMAP connection.

        #: Inbox folder, default to ```INBOX```.
        self.inbox = 'INBOX'
        #: Spam folder, default to ```INBOX.Spam```.
        self.spaminbox = 'INBOX.Spam'

        self.learnspambox = None     #: Folder used to learn spam messages.
        self.learnhambox = None      #: Folder used to learn non-spam messages.

        self._hashed_host = None
        self._hashed_user = None
        self._hashed_port = None
        self._hash = self.hash

    @property
    def hash(self):
        """str: Get the hash property built from the host, user and port."""
        if self._hashed_host != self.host or \
                self._hashed_user != self.user or \
                self._hashed_port != self.port:
            self._hashed_host = self.host
            self._hashed_user = self.user
            self._hashed_port = self.port
            self._hash = ImapSettings.get_hash(self.host, self.user, self.port)
        return self._hash

    @staticmethod
    def get_hash(host, user, port):
        """Get a hash with the host, user and port.

        Args:
            host (str): IMAP host name.
            user (str): IMAP user name.
            port (int): IMAP connection port.

        Returns:
            str: The generated hash.

        """
        newhash = md5()
        newhash.update(host.encode())
        newhash.update(user.encode())
        newhash.update(repr(port).encode())
        return newhash
